----------------------------------------------------------------------------------
-- Addon:         ProGuide - Guide Handler UI
-- Author:        Claus Gade Pedersen - Satisfyme
-- File:          ObjectiveFrameHandler.lua
-- Description:   Handles all the needed info to controll the Objective frame corectly.
----------------------------------------------------------------------------------
-- Addon version: 10
-- File Revision: 10
-- Last Edit:     2011-09-12T21:30:43Z
-- Last Edit By:  satisfyme
----------------------------------------------------------------------------------

----------------------------------------------------------------------------------
-- Local values
----------------------------------------------------------------------------------
local freeContentContainers, ContainerCount = {}, 0
local freeContentHeadlines, HeadlineCount = {}, 0
local freeContentImages, ImageCount = {}, 0
local freeContentItems, ItemCount = {}, 0
local freeContentCreateItems, CreateItemCount = {}, 0
local freeContentObjectives, ObjectiveCount = {}, 0
local freeContentTexts, TextCount = {}, 0
local freeContentTextHighlites, TextHighliteCount = {}, 0
local freeTables = {}
local L = ProGuideLocals
local objCount = 0
local offset = 0
local x, y = 0, 0
local RefreshFrames = {}
local Parent = {}
local tempFrames = {}
local sLevel, sID = 0, 0
local marker = ""
-- Content types allowed
local typeDef = {"all", "container", "headline", "image", "item", "createitem", "objective", "text", "highlite",}
local npcDef  = {"LearnFrom", "BuyFrom", "TrainFrom",}

-- Print Out Content Elements Used
function ProGuide:PrintContentElements()
  print("-------------------------------------")
  print("Element Number: Stored/Total")
  print("-------------------------------------")
  print("Containers: " .. tostring(#freeContentContainers) .. "/" .. tostring(ContainerCount))
  print("Headlines: " .. tostring(#freeContentHeadlines) .. "/" .. tostring(HeadlineCount))
  print("Images: " .. tostring(#freeContentImages) .. "/" .. tostring(ImageCount))
  print("Items: " .. tostring(#freeContentItems) .. "/" .. tostring(ItemCount))
  print("CreateItems: " .. tostring(#freeContentCreateItems) .. "/" .. tostring(CreateItemCount))
  print("Objectives: " .. tostring(#freeContentObjectives) .. "/" .. tostring(ObjectiveCount))
  print("Texts: " .. tostring(#freeContentTexts) .. "/" .. tostring(TextCount))
  print("TextHighlites: " .. tostring(#freeContentTextHighlites) .. "/" .. tostring(TextHighliteCount))
  print("-------------------------------------")
end

----------------------------------------------------------------------------------
-- local and some global functions
----------------------------------------------------------------------------------
-- Change objective icon
local function ChangeIcon(t, icon)
  if not t or type(t) ~= "table" then return end

  if ProGuide.Data.Icons[icon] then
    if type(ProGuide.Data.Icons[icon]) == "string" then
      t:SetTexture(ProGuide.Data.Icons[icon])
    elseif type(ProGuide.Data.Icons[icon]) == "table" then
      t:SetTexture(ProGuide.Data.Icons[icon].File)
      if ProGuide.Data.Icons[icon].Cords then
        if #ProGuide.Data.Icons[icon].Cords == 4 then
          t:SetTexCoord(
            ProGuide.Data.Icons[icon].Cords[1], 
            ProGuide.Data.Icons[icon].Cords[2],
            ProGuide.Data.Icons[icon].Cords[3],
            ProGuide.Data.Icons[icon].Cords[4]
          )
        elseif #ProGuide.Data.Icons[icon].Cords == 8 then
          t:SetTexCoord(
            ProGuide.Data.Icons[icon].Cords[1], 
            ProGuide.Data.Icons[icon].Cords[2],
            ProGuide.Data.Icons[icon].Cords[3],
            ProGuide.Data.Icons[icon].Cords[4],
            ProGuide.Data.Icons[icon].Cords[5],
            ProGuide.Data.Icons[icon].Cords[6],
            ProGuide.Data.Icons[icon].Cords[7],
            ProGuide.Data.Icons[icon].Cords[8]
          )
        end
      end
    else
      t:SetTexture(L["none"])
    end
  else
    t:SetTexture(L["none"])
  end
end

-- Get new/old content headline
local function GetContentContainer(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentContainers == 0 then ContainerCount = ContainerCount + 1 end
  return tremove(freeContentContainers) or CreateFrame("Frame", "ProGuideContainer" .. tostring(ContainerCount), Parent, "ProGuideContentContainerTemplate")
end

-- Get new/old content headline
local function GetContentHeadline(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentHeadlines == 0 then HeadlineCount = HeadlineCount + 1 end
  return tremove(freeContentHeadlines) or CreateFrame("Frame", "ProGuideHeadline" .. tostring(HeadlineCount), Parent, "ProGuideContentHeadlineTemplate")
end

-- Get new/old content image
local function GetContentImage(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentImages == 0 then ImageCount = ImageCount + 1 end
  return tremove(freeContentImages) or CreateFrame("Frame", "ProGuideImage" .. tostring(ImageCount), Parent, "ProGuideContentImageTemplate")
end

-- Get new/old content item
local function GetContentItem(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentItems == 0 then ItemCount = ItemCount + 1 end
  return tremove(freeContentItems) or CreateFrame("Frame", "ProGuideItem" .. tostring(ItemCount), Parent, "ProGuideContentItemTemplate")
end

-- Get new/old content create item
local function GetContentCreateItem(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentCreateItems == 0 then CreateItemCount = CreateItemCount + 1 end
  return tremove(freeContentCreateItems) or CreateFrame("Frame", "ProGuideCreateItem" .. tostring(CreateItemCount), Parent, "ProGuideContentCreateItemTemplate")
end

-- Get new/old content objective
local function GetContentObjective(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentObjectives == 0 then ObjectiveCount = ObjectiveCount + 1 end
  return tremove(freeContentObjectives) or CreateFrame("Frame", "ProGuideObjective" .. tostring(ObjectiveCount), Parent, "ProGuideContentObjectiveTemplate")
end

-- Get new/old content text
local function GetContentText(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentTexts == 0 then TextCount = TextCount + 1 end
  return tremove(freeContentTexts) or CreateFrame("Frame", "ProGuideText" .. tostring(TextCount), Parent, "ProGuideContentTextTemplate")
end

-- Get new/old content text
local function ContentTextHighlites(t, Parent)
  if not Parent then Parent = t.ScrollFrame.Content end
  if #freeContentTextHighlites == 0 then TextHighliteCount = TextHighliteCount + 1 end
  return tremove(freeContentTextHighlites) or CreateFrame("Frame", "ProGuideTextHighlite" .. tostring(TextHighliteCount), Parent, "ProGuideContentTextHighliteTemplate")
end

-- Get new/old table
function ProGuide:GetTable()
  return tremove(freeTables) or {}
end

-- Recycle old table
function ProGuide:RecycleTable(t)
  for key in pairs(t) do
    t[key] = nil
  end
  tinsert(freeTables, t)
end

-- Rearange Content
local function RearangeContent(t)
  for i = 1, #t.Content, 1 do
    if t.Content[i]["type"] == "text" then
      offset = 1
    else
      offset = -2
    end
    if i == 1 then
      t.Content[i].frame:SetPoint("TOP", t.ScrollFrame.Content, "TOP", 0, offset)
    else
      t.Content[i].frame:SetPoint("TOP", t.Content[i - 1].frame, "BOTTOM", 0, offset)
    end
  end
end

-- Racalculate content height
local function RecalculateContentHeight(t)
  t.ScrollFrame.Content:SetHeight(5)
  for i = 1, #t.Content, 1 do
    if t.Content[i]["type"] == "item" then
      if t.Content[i-1] then
        if t.Content[i-1]["type"] == "item" then
          offset = -2
        else
          offset = 0
        end
      else
        offset = 1
      end
    elseif t.Content[i]["type"] == "text" then
      offset = -1
    else
      offset = 2
    end
    t.ScrollFrame.Content:SetHeight(t.ScrollFrame.Content:GetHeight() + t.Content[i].frame:GetHeight() + offset)
  end
end

-- Recycle old content
local function RecycleContent(t, Rtype)
  if Rtype and (not tContains(typeDef, Rtype)) then return end

  if (not Rtype) or Rtype == "all" then Rtype = t.type end

  if Rtype == "container" then
    t.frame.Text:SetText("")
    for i=#t.frame.Content, 1, -1 do
      RecycleContent(t.frame.Content[i], t.frame.Content[i].type)
    end
    ProGuide:RecycleTable(t.frame.Content)
    t.frame.Content = nil
    t.frame:SetHeight(12)
    t.frame.Highlight:SetVertexColor(1, 1, 1, 0.2)
    t.frame:Hide()
    tinsert(freeContentContainers, t.frame)

  elseif Rtype == "headline" then
    t.frame:Hide()
    t.frame.Headline:SetText(UNUSED)
    tinsert(freeContentHeadlines, t.frame)

  elseif Rtype == "image" then
    t.frame:Hide()
    tinsert(freeContentImages, t.frame)

  elseif Rtype == "item" then
    t.frame:Hide()
    t.frame.ButtonText:Hide()
    t.frame.ButtonText.Text:SetText(UNUSED)
    t.frame.Counter:Hide()
    t.frame.DoneMarker:Hide()
    if t.frame.ButtonText.npc then t.frame.ButtonText.npc = nil end
    if t.frame.ButtonText.tooltip then t.frame.ButtonText.tooltip = nil end
    t.frame.ItemButton.IconGlow:Hide()
    ChangeIcon(t.frame.ItemButton.IconTexture)
    t.frame.Text:Hide()
    t.frame.Text:SetText(L["Getting Item Info"])
    tinsert(freeContentItems, t.frame)

  elseif Rtype == "createitem" then
    t.frame:Hide()
    t.frame.Highlight:Hide()
    t.frame.Counter:Hide()
    t.frame.DoneMarker:Hide()
    ChangeIcon(t.frame.Icon)
    t.frame.Text:Hide()
    t.frame.Text:SetText(L["Getting Item Info"])
    tinsert(freeContentCreateItems, t.frame)

  elseif Rtype == "objective" then
    t.frame:Hide()
    t.frame.ButtonText:Hide()
    t.frame.Highlight:Hide()
    ChangeIcon(t.frame.Icon)
    if t.frame.ButtonText.npc then t.frame.ButtonText.npc = nil end
    if t.frame.ButtonText.tooltip then t.frame.ButtonText.tooltip = nil end
    t.frame.ButtonText.Text:SetText(UNUSED)
    t.frame.DoneMarker:Hide()
    t.frame.MapButton:Hide()
    if t.frame.MapButton.file then t.frame.MapButton.file = nil end
    if t.frame.MapButton.map then t.frame.MapButton.map = nil end
    if t.frame.MapButton.tooltip then t.frame.MapButton.tooltip = nil end
    if t.frame.MapButton.zone then t.frame.MapButton.zone = nil end
    t.frame.Text:Hide()
    t.frame.Text:SetText(UNUSED)
    tinsert(freeContentObjectives, t.frame)

  elseif Rtype == "text" then
    t.frame:Hide()
    t.frame.Text:SetText(UNUSED)
    tinsert(freeContentTexts, t.frame)

  elseif Rtype == "highlite" then
    t.frame:Hide()
    t.frame.TextHighlite:SetText(UNUSED)
    tinsert(freeContentTextHighlites, t.frame)
  end

  if t.ItemWatchListHash then
    ProGuide:RemoveFromItemWatchList(t.ItemWatchListHash)
    t.ItemWatchListHash = nil
  end

end

-- Wait for return value
function ProGuide:WaitForItemInfo(t)
  tinsert(RefreshFrames, t)
end

-- Add recieved item info
local function RecievedInfo(t)
  if t.type == "item" or t.type == "itemBuyList" then
    local name, link, quality, _, _, _, _, _, _, texture, _ = GetItemInfo(t.itemId)
    if name then
      if t.npcId and ProGuide.Data.Npc[t.npcId] then
        t.frame.ButtonText.Text:SetText(name)
        t.frame.ButtonText.tooltip = L["BuyFrom"]:format(ProGuide.Data.Npc[t.npcId]["Name" .. GetLocale()] or ProGuide.Data.Npc[t.npcId].Name)
        t.frame.ButtonText.npc = t.npcId
        t.frame.ButtonText:Show()
        t.frame.Text:Hide()
      else
        t.frame.Text:SetText(name)
        t.frame.Text:Show()
        t.frame.ButtonText:Hide()
      end
      t.frame.ItemButton.link = link
      t.frame.ItemButton.IconTexture:SetTexture(texture)
      t.frame.ItemButton.IconGlow:SetVertexColor(GetItemQualityColor(quality))
      t.frame.ItemButton.IconGlow:Show()
      if t.checked then
        t.frame.DoneMarker:Show()
      elseif t.number and type(t.number) == "number" then
        if t.type == "itemBuyList" then
          if t.number > GetItemCount(t.itemId, true) then
            t.number = t.number - GetItemCount(t.itemId, true)
          else
            t.number = 0
          end
          t.frame.Counter:SetText(t.number)
        else
          t.frame.Counter:SetText(L["%d/%d"]:format( GetItemCount(t.itemId, true), t.number))
        end
        t.frame.Counter:Show()
      end
      t.ItemWatchListHash = ProGuide:AddToItemWatchList(t.frame)
    else
      ProGuide:WaitForItemInfo(t)
    end
  elseif t.type == "createitem" then
    local name = GetItemInfo(t.itemId)
    if name then
      t.frame.Text:SetText(L["Create %s"]:format(name))
      t.frame.Text:Show()
      if checked then
        t.frame.DoneMarker:Show()
      else
        t.frame.Counter:SetText(L["%d/%d"]:format( GetItemCount(t.itemId, true), t.number))
        t.frame.Counter:Show()
      end
      t.ItemWatchListHash = ProGuide:AddToItemWatchList(t.frame)
    else
      ProGuide:WaitForItemInfo(t)
    end
  end
end

----------------------------------------------------------------------------------
-- Update Item Info by using Event: GET_ITEM_INFO_RECEIVED
----------------------------------------------------------------------------------
local infoCatchFrame = CreateFrame("Frame")
infoCatchFrame:SetScript("OnEvent", function(t, event)
  if #RefreshFrames > 0 then
    -- Copy anc clear RefreshFrame
    tempFrames = RefreshFrames
    RefreshFrames = {}
    for i = #tempFrames, 1, -1 do
      RecievedInfo(tempFrames[i])
    end
    tempFrames = {}
  end
end)
infoCatchFrame:RegisterEvent("GET_ITEM_INFO_RECEIVED")

----------------------------------------------------------------------------------
-- Global functions
----------------------------------------------------------------------------------
-- Create a new objective frame
function ProGuide:ObjInitialice()
  -- Calculate new obj count
  objCount = objCount + 1

  -- Create a new Objective frame
  local obj = CreateFrame("Frame", "ProGuideObj" .. tostring(objCount), UIParent, "ProGuideObjFramesTemplate")
  -- we only needs the first one to be moveable
  if objCount > 1 then
    obj:SetPoint("TOP", "ProGuideObj" .. tostring(objCount - 1), "BOTTOM" ,0 , 10)
    obj:SetMovable(false)
    obj:RegisterForDrag("None")
  end
  obj.ButtonGuides:SetText(L["Guides"])
  obj.ButtonBuyList:SetText(L["Buy List"])
  obj.ButtonBuyList.tooltip = L["BuyListTooltip"]
  obj.ButtonBuyList.guide = false
  obj.ButtonBuyList.objID = objCount
  obj.ButtonBuyList.part = false

  obj.Map = CreateFrame("Button", "ProGuideMap" .. tostring(objCount), obj, "ProGuideMapFramesTemplate")
  obj.Map:SetPoint("RIGHT", obj, "LEFT")

  obj.BuyList = CreateFrame("Frame", "ProGuideBuyList" .. tostring(objCount), obj, "ProGuideBuyListFrameTemplate")
  obj.BuyList:SetPoint("TOPRIGHT", obj, "TOPLEFT")
  obj.BuyList.ButtonClose:SetText(CLOSE)
  obj.BuyList.Headline:SetText(L["Buy List"])
  obj.BuyList.guide = false
  obj.BuyList.objID = objCount
  obj.BuyList.part = false

  obj.ButtonBuyList.BuyList = obj.BuyList

  obj.Content = ProGuide:GetTable()

  -- Add default headline
  ProGuide:EditHeadline(obj, L["No Guide Loaded"])
  --[[ProGuide:EditSubHeadline(obj, "Skill 200-300")
  ProGuide:EditTextLevel(obj, "Skill Level 245")
  ProGuide:AddContentHeadline(obj, "Content Headline 01")
  --ProGuide:AddContentObjective(t, icon, mark, text, npcId, npcType, checked, map, tooltip, zone, id)
  --npcType  = {"LearnFrom", "BuyFrom", "TrainFrom",}
  ProGuide:AddGetList(obj, {
    [1] = {
      ["Item"] = 29739,
      ["Num"] = 100,
      ["BuyFrom"] = 1347,
    },
    [2] = {
      ["Item"] = 49482,
      ["Num"] = 1,
    },
  }, nil)
  ProGuide:AddContentObjective(obj, nil, nil, "With Map", nil, nil, true, "A_Herbalism_ElwynnForest.tga", "Herbalism Grinding Zone", "Elwynn Forest")
  ProGuide:AddContentObjective(obj, nil, true, "With Map but no zone looooong text", nil, nil, nil, "A_Mining_ElwynnForest.tga", "Mining Grinding Zone", nil)
  ProGuide:AddContentObjective(obj, nil, nil, "Without Map")
  ProGuide:AddContentObjective(obj, "teleport", nil, "Train Alchemy", 1347, "TrainFrom")
  ProGuide:AddContentObjective(obj, nil, nil, "Compleated", nil, nil, true)
  ProGuide:AddContentText(obj, "Content text that holds large textareas with multiline of some bla bla bla bla bla bla bla bla bla bla bla bla bla bla bla")
  ProGuide:AddContentCreateItem(obj, nil, true, 21840, 200, nil, 1346)
  ProGuide:AddContentCreateItem(obj, nil, nil, 21840, 200, true, 1346)
  ProGuide:AddContentImage(obj, "Interface\\AddOns\\TourGuide-Tradeskill\\Maps\\H_Herbalism_Durotar.tga", 384, 256, nil, 0.75, nil, nil, nil)
  ]]
  -- return the new Objective frame
  return obj
end

-- Add Content Get List
function ProGuide:AddGetList(t, items, mark, text, id)
  if not t or type(t) ~= "table" then return end
  if not items or type(items) ~= "table" then return end
  if not text or type(text) ~= "string" or strlen(text) <= 0 then text = L["Get items listed below"] end
  if not id or type(id) ~= "number" then id = #t.Content + 1 end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentContainer(t),
    ["type"] = "container",})

  t.Content[id].frame.Text:SetText(text)

  -- Add items to the content
  if not t.Content[id].frame.Content then t.Content[id].frame.Content = ProGuide:GetTable() end
  for key, val in ipairs(items) do
    ProGuide:AddContentItem(t.Content[id].frame, val.Item, val.Num, nil, val.BuyFrom, nil, t.Content[id].frame)
    if key == 1 then
      t.Content[id].frame.Content[key].frame:SetPoint("TOP", t.Content[id].frame.Text, "BOTTOM", 0, 0)
    else
      t.Content[id].frame.Content[key].frame:SetPoint("TOP", t.Content[id].frame.Content[key - 1].frame, "BOTTOM", 0, 0)
    end
  end

  if mark then
    t.Content[id].frame.Highlight:SetVertexColor(1, 0.82, 0, 0.5)
  else
    t.Content[id].frame.Highlight:SetVertexColor(1, 1, 1, 0.2)
  end

  -- Set Height
  t.Content[id].frame:SetHeight(16 * #items + 12)

  -- Show Get List
  t.Content[id].frame:Show()

  -- Finalice
  RearangeContent(t)
  RecalculateContentHeight(t)
end

-- Add Note To Objective
local NoteTypes = {"objective",}
function ProGuide:AddNoteToObjective(t, note, info, id)
  if not t or type(t) ~= "table" then return end
  if not note or type(note) ~= "string" or strlen(note) <= 0 then return end
  if not tContains(NoteTypes, info[1]) then return end
  if not id or type(id) ~= "number" then id = #t.Content + 1 end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentContainer(t),
    ["type"] = "container",
  })

  t.Content[id].frame.Text:SetText("")

  -- Add items to the content
  if not t.Content[id].frame.Content then t.Content[id].frame.Content = ProGuide:GetTable() end

  -- Add original content
  if info[1] == "objective" then
    ProGuide:AddContentObjective(t.Content[id].frame, info[2], info[3], info[4], info[5], info[6], info[7], info[8], info[9], info[10], nil, t.Content[id].frame)
  end
  t.Content[id].frame.Content[1].frame:SetPoint("TOP", t.Content[id].frame, "TOP", 0, 0)

  -- Add note
  ProGuide:AddContentText(t.Content[id].frame, note, t.Content[id].frame)
  t.Content[id].frame.Content[2].frame:SetPoint("TOP", t.Content[id].frame.Content[1].frame, "BOTTOM", 0, 0)

  if mark then
    t.Content[id].frame.Highlight:SetVertexColor(1, 0.82, 0, 0.5)
  else
    t.Content[id].frame.Highlight:SetVertexColor(1, 1, 1, 0.2)
  end

  -- Set Height
  t.Content[id].frame:SetHeight(t.Content[id].frame.Content[1].frame:GetHeight() + t.Content[id].frame.Content[2].frame:GetHeight())

  -- Show Get List
  t.Content[id].frame:Show()

  -- Finalice
  RearangeContent(t)
  RecalculateContentHeight(t)
end

-- Edit Headline
function ProGuide:EditHeadline(t, text)
  if (not t) or type(t) ~= "table" then return end
  if not text then text = "" end
  t.TextHeadline:SetText(text)
end

-- Edit Sub Headline
function ProGuide:EditSubHeadline(t, text)
  if (not t) or type(t) ~= "table" then return end
  if not text then text = "" end
  t.TextSubHeadline:SetText(text)
end

-- Edit Level
function ProGuide:EditTextLevel(t, text)
  if (not t) or type(t) ~= "table" then return end
  if not text then text = "" end
  t.TextLevel:SetText(text)
end

-- Edit Icon
function ProGuide:EditIcon(t, link)
  if not t or type(t) ~= "table" or not t.Icon then return end

  if link and type(link) == "string" then
    -- Set icon
    SetPortraitToTexture(t.Icon, link)
  else
    return
  end
end

-- Set Skill Level
function ProGuide:SetSkillLevel(t, skillID)
  if not skillID or type(skillID) ~= "number" then return end
  
  -- Run thrue the players professions to check if they have the skill
  for i=1, 6 do
    sID = select(i, GetProfessions())
    -- if profession is found, then show it
    if sID and select(7,GetProfessionInfo(sID)) == skillID then
      ProGuide:EditTextLevel(t, L["%s level: %d"]:format(ProGuide.Data.PN[skillID], select(3, GetProfessionInfo(sID))))
      return
    end
  end
  
  -- Did not find a skill by the id, then we don't have it and print out skill level 0
  ProGuide:EditTextLevel(t, L["%s level: %d"]:format(ProGuide.Data.PN[skillID], 0))
end

-- Add Content Headline
function ProGuide:AddContentHeadline(t, text, id)
  if (not t) or type(t) ~= "table" then return end
  if (not text) or text == "" then return end
  if (not id) or type(id) ~= "number" then id = #t.Content + 1 end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentHeadline(t),
    ["type"] = "headline",
  })
  
  t.Content[id].frame.Headline:SetText(text)
  t.Content[id].frame:SetHeight(t.Content[id].frame.Headline:GetStringHeight() + 2)
  t.Content[id].frame:Show()
  RearangeContent(t)
  RecalculateContentHeight(t)
end

-- Add Content Image
function ProGuide:AddContentImage(t, file, width, height, left, right, top, bottom, id)
  if (not t) or type(t) ~= "table" then return end
  if (not file) or file == "" then return end
  if (not id) or type(id) ~= "number" then id = #t.Content + 1 end
  if (not width) or (not height) or type(width) ~= "number" or type(height) ~= "number" then return end
  if width <= 270 then
    x = width
    y = height
  else
    x = 270
    y = (270 / width) * height
  end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentImage(t),
    ["type"] = "image",})
  
  t.Content[id].frame:SetSize(x, y)
  t.Content[id].frame.Image:SetTexture(file)
  t.Content[id].frame.Image:SetSize(x, y)
  t.Content[id].frame.Image:SetTexCoord(left or 0, right or 1, top or 0, bottom or 1)
  t.Content[id].frame:Show()
  RearangeContent(t)
  RecalculateContentHeight(t)
end

-- Add Content Item
function ProGuide:AddContentItem(t, itemId, number, checked, npcId, mark, Parent, id)
  if (not t) or type(t) ~= "table" then return end
  if (not itemId) or type(itemId) ~= "number" then return end
  if (not id) or type(id) ~= "number" then id = #t.Content + 1 end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentItem(t, Parent),
    ["itemId"] = itemId,
    ["type"] = "item",
    ["number"] = number,
  })

  if mark then
    t.Content[id].frame.Highlight:Show()
  else
    t.Content[id].frame.Highlight:Hide()
  end
  
  local name, link, quality, _, _, _, _, _, _, texture, _ = GetItemInfo(itemId)
  if name then
    if npcId and ProGuide.Data.Npc[npcId] then
      t.Content[id].frame.ButtonText.Text:SetText(name)
      t.Content[id].frame.ButtonText.tooltip = L["BuyFrom"]:format(ProGuide.Data.Npc[npcId]["Name" .. GetLocale()] or ProGuide.Data.Npc[npcId].Name)
      t.Content[id].frame.ButtonText.npc = npcId
      t.Content[id].frame.ButtonText:Show()
      t.Content[id].frame.Text:Hide()
    else
      t.Content[id].frame.Text:SetText(name)
      t.Content[id].frame.Text:Show()
      t.Content[id].frame.ButtonText:Hide()
    end
    t.Content[id].frame.ItemButton.link = link
    t.Content[id].frame.ItemButton.IconTexture:SetTexture(texture)
    t.Content[id].frame.ItemButton.IconGlow:SetVertexColor(GetItemQualityColor(quality))
    t.Content[id].frame.ItemButton.IconGlow:Show()
    if checked then
      t.Content[id].frame.DoneMarker:Show()
    elseif number and type(number) == "number" then
      t.Content[id].frame.Counter:SetText(L["%d/%d"]:format( GetItemCount(itemId, true), number))
      t.Content[id].frame.Counter:Show()
    end
    t.Content[id].ItemWatchListHash = ProGuide:AddToItemWatchList(t.Content[id])
  else
    t.Content[id].frame.Text:SetText(L["Getting Item Info"])
    t.Content[id].frame.Text:Show()
    ChangeIcon(t.Content[id].frame.ItemButton.IconTexture)
    t.Content[id].checked = checked
    t.Content[id].npcId = npcId
    ProGuide:WaitForItemInfo(t.Content[id])
  end

  t.Content[id].frame:Show()
  
  if not Parent then
    RearangeContent(t)
    RecalculateContentHeight(t)
  end
end

-- Add Content Item
function ProGuide:AddContentCreateItem(t, icon, mark, itemId, number, checked, npcId, id)
  if (not t) or type(t) ~= "table" then return end
  if (not itemId) or type(itemId) ~= "number" then return end
  if (not number) or type(number) ~= "number" then return end
  if (not id) or type(id) ~= "number" then id = #t.Content + 1 end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentCreateItem(t),
    ["itemId"] = itemId,
    ["type"] = "createitem",
    ["number"] = number,
  })
  
  -- change icon if avalible
  if not icon then icon = "craft" end
  ChangeIcon(t.Content[id].frame.Icon, icon)

  if mark then
    t.Content[id].frame.Highlight:Show()
  else
    t.Content[id].frame.Highlight:Hide()
  end

  local name = GetItemInfo(itemId)
  if name then
    t.Content[id].frame.Text:SetText(L["Create %s"]:format(name))
    t.Content[id].frame.Text:Show()
    if checked then
      t.Content[id].frame.DoneMarker:Show()
    else
      t.Content[id].frame.Counter:SetText(L["%d/%d"]:format( GetItemCount(itemId, true), number))
      t.Content[id].frame.Counter:Show()
    end
    t.Content[id].ItemWatchListHash = ProGuide:AddToItemWatchList(t.Content[id])
  else
    t.Content[id].frame.Text:SetText(L["Getting Item Info"])
    t.Content[id].frame.Text:Show()
    t.Content[id].number = number
    t.Content[id].checked = checked
    t.Content[id].npcId = npcId
    ProGuide:WaitForItemInfo(t.Content[id])
  end

  t.Content[id].frame:Show()
  RearangeContent(t)
  RecalculateContentHeight(t)
end

-- Add Content Objective
function ProGuide:AddContentObjective(t, icon, mark, text, npcId, npcType, checked, map, mapinfo, zone, note, Parent, id)
  if (not t) or type(t) ~= "table" then return end
  if (not text) or text == "" then return end
  if (npcId and (not npcType)) then return end
  if npcId and (type(npcId) ~= "number" or (not tContains(npcDef, npcType))) then return end
  if note then 
    ProGuide:AddNoteToObjective(t, note, {"objective", icon, mark, text, npcId, npcType, checked, map, mapinfo, zone,}, id)
    return
  end
  if (not id) or type(id) ~= "number" then id = #t.Content + 1 end


  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentObjective(t, Parent),
    ["type"] = "objective",
  })
  
  -- change icon if avalible
  ChangeIcon(t.Content[id].frame.Icon, icon)
  
  if mark then 
    t.Content[id].frame.Highlight:Show()
  else
    t.Content[id].frame.Highlight:Hide()
  end

  if npcId and ProGuide.Data.Npc[npcId] then
    t.Content[id].frame.ButtonText.Text:SetText(text)
    t.Content[id].frame.ButtonText.tooltip = L[npcType]:format(ProGuide.Data.Npc[npcId]["Name" .. GetLocale()] or ProGuide.Data.Npc[npcId].Name)
    t.Content[id].frame.ButtonText.npc = npcId
    t.Content[id].frame.ButtonText:Show()
    t.Content[id].frame.Text:Hide()
  else
    t.Content[id].frame.Text:SetText(text)
    t.Content[id].frame.Text:Show()
    t.Content[id].frame.ButtonText:Hide()
  end
  t.Content[id].frame.Text:SetText(text)
  if checked then
    t.Content[id].frame.DoneMarker:Show()
  elseif map then
    t.Content[id].frame.MapButton.map = t.Map
    t.Content[id].frame.MapButton.file = map
    t.Content[id].frame.MapButton.tooltip = L["ClickToShowMap"]:format(mapinfo)
    t.Content[id].frame.MapButton.zone = zone
    t.Content[id].frame.MapButton:Show()
  end
  t.Content[id].frame:Show()
  if not Parent then
    RearangeContent(t)
    RecalculateContentHeight(t)
  end
end

-- Add Content Text
function ProGuide:AddContentText(t, text, Parent, id)
  if (not t) or type(t) ~= "table" then return end
  if (not text) or text == "" then return end
  if (not id) or type(id) ~= "number" then id = #t.Content + 1 end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = GetContentText(t, Parent),
    ["type"] = "text",
  })
  
  t.Content[id].frame.Text:SetText(text)
  t.Content[id].frame:SetHeight(t.Content[id].frame.Text:GetStringHeight() + 2)
  t.Content[id].frame:Show()
  if not Parent then
    RearangeContent(t)
    RecalculateContentHeight(t)
  end
end                                      

-- Add Content Text Highlite
function ProGuide:AddContentTextHighlite(t, text, id)
  if (not t) or type(t) ~= "table" then return end
  if (not text) or text == "" then return end
  if (not id) or type(id) ~= "number" then id = #t.Content + 1 end

  tinsert(t.Content, {
    ["id"] = id,
    ["part"] = t.Part,
    ["guidename"] = t.GuideName,
    ["frame"] = ContentTextHighlites(t),
    ["type"] = "highlite",})
  
  t.Content[id].frame.TextHighlite:SetText(text)
  t.Content[id].frame:SetHeight(t.Content[id].frame.TextHighlite:GetStringHeight() + 2)
  t.Content[id].frame:Show()
  RearangeContent(t)
  RecalculateContentHeight(t)
end

-- Remove Content
function ProGuide:RemoveContent(t, Rtype, pos, id)
  if (not t) or type(t) ~= "table" then return end
  if #t.Content <= 0 then return end
  if ((not Rtype) or (not tContains(typeDef, Rtype))) and
     ((not pos) or type(pos) ~= "number") and
     ((not id) or type(id) ~= "number")
  then
    return
  end

  if Rtype and pos then
    if t.Content[pos].type == Rtype then
      RecycleContent(t.Content[pos], Rtype)
      tremove(t.Content, pos)
    end
  elseif Rtype and id then
    for i = #t.Content, 1, -1 do
      if t.Content[i].id == id then
        if t.Content[i].type == Rtype then
          RecycleContent(t.Content[i], Rtype)
          tremove(t.Content, i)
        end
      end
    end
  elseif pos then
    RecycleContent(t.Content[pos])
    tremove(t.Content, pos)
  elseif id then
    for i = #t.Content, 1, -1 do
      if t.Content[i].id == id then
        RecycleContent(t.Content[i])
        tremove(t.Content, i)
      end
    end
  else
    for i = #t.Content, 1, -1 do
      if Rtype == "all" or Rtype == t.Content[i].type then
        RecycleContent(t.Content[i], Rtype)
        tremove(t.Content, i)
      end
    end
  end
  RearangeContent(t)
  RecalculateContentHeight(t)
end

-- Add Learn Item to objective
function ProGuide:AddCreateItem(obj, mark, info, done)
  if not obj then return end
  if not info or type(info) ~= "table" then return end

  ProGuide:AddContentCreateItem(
    obj,
    "craft",
    mark,
    info.Item, 
    info.Num,
    done
  )
end

-- Add Learn Item to objective
function ProGuide:AddLearnItem(obj, mark, info, done)
  if not obj then return end
  if not info or type(info) ~= "table" then return end

  ProGuide:AddContentObjective(
    obj,
    "learn",
    mark,
    L["Learn %s%s"]:format(ProGuide.Data.ItemType[info.Skill], ProGuide.Data.Items[info.ItemID]),
    info.LearnFrom,
    "LearnFrom",
    done
  )
end

-- Add Train Skill to objective
function ProGuide:AddTrainSkill(obj, mark, info, done)
  if not obj then return end
  if not info or type(info) ~= "table" then return end
  --AddContentObjective(t, icon, mark, text, npcId, npcType, checked, map, tooltip, zone, id)
  ProGuide:AddContentObjective(
    obj,
    "train",
    mark,
    L["Train %s %s"]:format(ProGuide.Data.PLN[info.Level], ProGuide.Data.PNT[info.Skill]),
    info.TrainFrom,
    "TrainFrom",
    done
  )
end

-- Add Train Skill to objective
function ProGuide:AddLevelSkill(obj, mark, info, done)
  if not obj then return end
  if not info or type(info) ~= "table" then return end
  
  --AddContentObjective(t, icon, mark, text, npcId, npcType, checked, map, mapinfo, zone, tooltip, id)
  if info.Create then
    info.Tooltip = L["Create "]
    info.EndText = ""
    for i=1, #info.Create do
      if i > 1 then marker = ", " else marker = "" end
      if i == #info.Create then
        if info.CreateEndNote and type(info.CreateEndNote) == "string" and strlen(info.CreateEndNote) > 0 then
          info.EndText = ", " .. info.CreateEndNote .. "."
        else
          info.EndText = "."
        end
        if i > 1 then marker = L[" or "] else marker = "" end
      end
      info.Tooltip = info.Tooltip .. marker .. ProGuide.Data.Items[info.Create[i].Item] .. info.EndText
    end
  end

  ProGuide:AddContentObjective(
    obj,
    "levelskill",
    mark,
    L["Level %s to %d"]:format(ProGuide.Data.PN[info.Skill], info.Level),
    nil,
    nil,
    done,
    nil,
    nil,
    nil,
    info.Tooltip
  )
end

